/* Distributed Checksum Clearinghouse
 *
 * continually restart a daemon
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.5-1.7 $Revision$
 */

#include <sys/wait.h>
#include <signal.h>
#include "dcc_clnt.h"

#define MIN_RESTART_DELAY   5
#define MAX_RESTART_DELAY   (5*60)

/* stay alive despite core dumps
 *	Restart immediately, but not more often than every MINRESTART_DELAY
 *	seconds.  If the restarts are happening at the maximum rate,
 *	halve the rate. */
void
dcc_daemon_restart(SOCKET sock_fd, void(reconn)(void))
{
#if defined(WIFEXITED) && defined(WTERMSIG) && defined(WIFSIGNALED)
	pid_t start_pid, pid;
	time_t next_restart, restart_delay;
	int status;

	restart_delay = MIN_RESTART_DELAY;
	next_restart = 0;
	for (;;) {
		start_pid = fork();
		if (!start_pid) {
			/* reconnect sockets or whatever except first time */
			if (next_restart != 0
			    && reconn)
				reconn();
			return;
		}

		if (start_pid < 0)
			dcc_logbad(EX_OSERR, "(re)start fork(): %s",
				   ERROR_STR());

		if (sock_fd != -1) {
			close(sock_fd);
			sock_fd = -1;
		}

		next_restart = time(0)+restart_delay;
		pid = waitpid(start_pid, &status, 0);
		if (pid < 0) {
			if (errno != EINTR)
				dcc_logbad(EX_OSERR, "restart waitpid(): %s",
					   ERROR_STR());
			exit(0);
		}
		if (WIFEXITED(status)) {
			if (dcc_clnt_debug)
				dcc_error_msg("do not restart after exit %d",
					      WEXITSTATUS(status));
			exit(WEXITSTATUS(status));
		}
		if (WIFSIGNALED(status)
		    && (WTERMSIG(status) != SIGILL
			&& WTERMSIG(status) != SIGSEGV
#ifdef SIGBUS
			&& WTERMSIG(status) != SIGBUS
#endif
#ifdef SIGXCPU
			&& WTERMSIG(status) != SIGXCPU
#endif
#ifdef SIGFPE
			&& WTERMSIG(status) != SIGFPE
#endif
#ifdef SIGSYS
			&& WTERMSIG(status) != SIGSYS
#endif
#ifdef SIGTRAP
			&& WTERMSIG(status) != SIGTRAP
#endif
			&& WTERMSIG(status) != SIGQUIT
			&& WTERMSIG(status) != SIGABRT)) {
			if (dcc_clnt_debug)
				dcc_error_msg("do not restart after signal #%d",
					      WTERMSIG(status));
			exit(0);
		}
		next_restart -= time(0);
		if (next_restart > 0 && next_restart <= MAX_RESTART_DELAY) {
			restart_delay *= 2;
			if (restart_delay > MAX_RESTART_DELAY)
				restart_delay = MAX_RESTART_DELAY;
			sleep(next_restart);
		} else {
			restart_delay -= next_restart-1;
			if (restart_delay < MIN_RESTART_DELAY)
				restart_delay = MIN_RESTART_DELAY;
		}
		dcc_error_msg("restart after signal #%d",
			      WTERMSIG(status));
	}
#endif /* not POSIX */
}
