/* Distributed Checksum Clearinghouse
 *
 * SMTP reply strings
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.5-1.8 $Revision$
 */


#include "cmn_defs.h"

REPLY rej_reply;
REPLY grey_reply;
REPLY temp_reply;


void
parse_reply_arg(const char *arg)
{
	if (rej_reply.rcode[0] == '\0') {
		parse_reply(&rej_reply, 1,
			    DCC_XCODE, DCC_RCODE, arg);
	} else if (grey_reply.rcode[0] == '\0') {
		parse_reply(&grey_reply, 1,
			    GREY_XCODE, GREY_RCODE, arg);
		if (grey_reply.rcode[0] != '4'
		    || grey_reply.xcode[0] != '4') {
			dcc_error_msg("invalid greylist"
				      " message: %s", arg);
			grey_reply.rcode[0] = '\0';
		}
	} else {
		dcc_error_msg("more than 2 -r settings");
	}
}



void
finish_replies(void)
{
	parse_reply(&temp_reply, 0, "4.7.1", "451", "DCC failure");
	if (rej_reply.rcode[0] == '\0')
		parse_reply(&rej_reply, 1, DCC_XCODE, DCC_RCODE,
			    "mail %s from %s rejected by DCC");
	if (grey_reply.rcode[0] == '\0')
		parse_reply(&grey_reply, 1, GREY_XCODE, GREY_RCODE,
			    "mail %s from %s temporary greylist embargoed");
}



/* Parse a string into RFC 821 and RFC 2034 codes and a pattern to make a
 * message, or generate the codes for a string that has a message
 * without codes.
 * The should be something like "5.7.1 550 spammer go home"
 * or "spammer go home".
 * The pattern can have as many as two "%s" references to the sendmail
 * queue-ID and SMTP client IP address.
 */
void
parse_reply(REPLY *reply,		/* to here */
	    u_char pat_ok,
	    const char *xcode,		/* default value such as "5.7.1" */
	    const char *rcode,		/* default value such as "550" */
	    const char *str)		/* from here */
{
	const char *p;
	char c, *pat;
	char sb[5+1+3+1];
	int p_cnt, i;

	/* Does the message have a valid xcode and rcode?
	 * First check for sendmail.cf ERROR:### and ERROR:D.S.N:### */
	if (!pat_ok && !CSTRCMP(str, "ERROR:")) {
		p = dcc_parse_word(0, sb, sizeof(sb), str+STRZ("ERROR:"),
				   0, 0, 0);
		if (p) {
			i = strlen(sb);
			if (i == 5+1+3 && sb[5] == ':') {
				memcpy(reply->xcode, &sb[0], 5);
				memcpy(reply->rcode, &sb[6], 3);
			} else if (i == 3) {
				BUFCPY(reply->xcode, xcode);
				memcpy(reply->rcode, &sb[0], 3);
			} else {
				p = 0;
			}
		}
	} else {
		p = dcc_parse_word(0, reply->xcode, sizeof(reply->xcode), str,
				   0, 0, 0);
		p = dcc_parse_word(0, reply->rcode, sizeof(reply->rcode), p,
				   0, 0, 0);
	}
	if (!p
	    || reply->rcode[0] < '4' || reply->rcode[0] > '5'
	    || reply->rcode[1] < '0' || reply->rcode[1] > '9'
	    || reply->rcode[2] < '0' || reply->rcode[2] > '9'
	    || reply->rcode[0] != reply->xcode[0]
	    || reply->xcode[1] != '.'
	    || reply->xcode[2] < '0' || reply->xcode[2] > '9'
	    || reply->xcode[3] != '.'
	    || reply->xcode[4] < '0' || reply->xcode[4] > '9') {
		BUFCPY(reply->xcode, xcode);
		BUFCPY(reply->rcode, rcode);
		p = str;
	}

	p += strspn(p, DCC_WHITESPACE);
	if (!pat_ok || !strchr(p, '%')) {
		reply->is_pat = 0;
		BUFCPY(reply->pat, p);
	} else {
		reply->is_pat = 1;
		p_cnt = 0;
		pat = reply->pat;
		do {
			c = *p++;
			if (c == '\0')
				break;
			*pat++ = c;
			if (c == '%') {
				if (*p != 's' || ++p_cnt > 2)
					*pat++ = '%';
			}
		} while (pat < LAST(reply->pat)-2);
		*pat = '\0';
	}
}



void
make_reply(CMN_WORK *cwp, REPLY *reply)
{
	const char *clnt;

	memcpy(cwp->reply.rcode, reply->rcode, sizeof(cwp->reply.rcode));
	memcpy(cwp->reply.xcode, reply->xcode, sizeof(cwp->reply.xcode));
	if (!reply->is_pat) {
		cwp->reply_ptr = reply->pat;
	} else {
		clnt = cwp->clnt_str;
		if (!strncmp("::ffff:", clnt, 7))
			clnt += 7;	/* strip IPv6 prefix */
		snprintf(cwp->reply_buf, sizeof(cwp->reply_buf),
			 reply->pat, cwp->id, clnt);
		cwp->reply_ptr = cwp->reply_buf;
	}
}
